from typing import List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import WaterfallStep


def translate_waterfall(step: WaterfallStep) -> List[MongoStep]:
    concatMongo = {}
    facet = {}
    groupby = step.groupby
    parents = [step.parentsColumn] if step.parentsColumn else []

    def _column_map(colnames: list) -> dict:
        return {col: f'${col}' for col in colnames}

    # Pipeline that will be executed to get the array of results for the starting
    # and ending milestone of the waterfall
    facetStartEnd = [
        {
            '$group': {
                '_id': _column_map(groupby + [step.milestonesColumn]),
                step.valueColumn: {'$sum': f'${step.valueColumn}'},
            },
        },
        {
            '$project': {
                **{col: f'$_id.{col}' for col in groupby},
                'LABEL_waterfall': f'$_id.{step.milestonesColumn}',
                **(
                    {'GROUP_waterfall': f'$_id.{step.milestonesColumn}'}
                    if step.parentsColumn
                    else {}
                ),
                'TYPE_waterfall': None,
                step.valueColumn: 1,
                '_vqbOrder': {
                    '$cond': [{'$eq': [f'$_id.{step.milestonesColumn}', step.start]}, -1, 1]
                },
            },
        },
    ]

    # Pipeline that will be executed to get the array of results for the children
    # elements of the waterfall
    facetChildren = [
        {
            '$group': {
                '_id': _column_map(
                    groupby + parents + [step.labelsColumn] + [step.milestonesColumn]
                ),
                step.valueColumn: {'$sum': f'${step.valueColumn}'},
            },
        },
        {
            '$addFields': {
                '_vqbOrder': {
                    '$cond': [{'$eq': [f'$_id.{step.milestonesColumn}', step.start]}, 1, 2]
                },
            },
        },
        {'$sort': {'_vqbOrder': 1}},
        {
            '$group': {
                '_id': {col: f'$_id.{col}' for col in groupby + parents + [step.labelsColumn]},
                '_vqbValuesArray': {'$push': f'${step.valueColumn}'},
            },
        },
        {
            '$project': {
                **{col: f'$_id.{col}' for col in groupby},
                'LABEL_waterfall': f'$_id.{step.labelsColumn}',
                **({'GROUP_waterfall': f'$_id.{step.parentsColumn}'} if step.parentsColumn else {}),
                'TYPE_waterfall': 'child' if step.parentsColumn else 'parent',
                step.valueColumn: {
                    '$reduce': {
                        'input': '$_vqbValuesArray',
                        'initialValue': 0,
                        'in': {'$subtract': ['$$this', '$$value']},
                    },
                },
                '_vqbOrder': {'$literal': 0},
            },
        },
    ]

    # If.parentsColumn is define, we set the pipeline that will be executed to
    # get the array of results for the parents elements of the waterfall. In such
    # a case we add it to the concatenation of all the pipelines results arrays
    if step.parentsColumn:
        facetParents = [
            {
                '$group': {
                    '_id': _column_map(groupby + parents + [step.milestonesColumn]),
                    step.valueColumn: {'$sum': f'${step.valueColumn}'},
                },
            },
            {
                '$addFields': {
                    '_vqbOrder': {
                        '$cond': [{'$eq': [f'$_id.{step.milestonesColumn}', step.start]}, 1, 2]
                    },
                },
            },
            {'$sort': {'_vqbOrder': 1}},
            {
                '$group': {
                    '_id': {col: f'$_id.{col}' for col in groupby + parents},
                    '_vqbValuesArray': {'$push': f'${step.valueColumn}'},
                },
            },
            {
                '$project': {
                    **{col: f'$_id.{col}' for col in groupby},
                    'LABEL_waterfall': f'$_id.{step.parentsColumn}',
                    'GROUP_waterfall': f'$_id.{step.parentsColumn}',
                    'TYPE_waterfall': 'parent',
                    (step.valueColumn): {
                        '$reduce': {
                            'input': '$_vqbValuesArray',
                            'initialValue': 0,
                            'in': {'$subtract': ['$$this', '$$value']},
                        },
                    },
                    '_vqbOrder': {'$literal': 0},
                },
            },
        ]

        facet = {
            '_vqb_start_end': facetStartEnd,
            '_vqb_parents': facetParents,
            '_vqb_children': facetChildren,
        }
        concatMongo = {'$concatArrays': ['$_vqb_start_end', '$_vqb_parents', '$_vqb_children']}
    else:
        facet = {'_vqb_start_end': facetStartEnd, '_vqb_children': facetChildren}
        concatMongo = {'$concatArrays': ['$_vqb_start_end', '$_vqb_children']}

    return [
        {'$match': {(step.milestonesColumn): {'$in': [step.start, step.end]}}},
        {'$facet': facet},
        {'$project': {'_vqbFullArray': concatMongo}},
        {'$unwind': '$_vqbFullArray'},
        {'$replaceRoot': {'newRoot': '$_vqbFullArray'}},
        {
            '$sort': {
                '_vqbOrder': 1,
                ('LABEL_waterfall' if step.sortBy == 'label' else step.valueColumn): 1
                if step.order == 'asc'
                else -1,
            },
        },
        {'$project': {'_vqbOrder': 0}},
    ]



from typing import Any, Dict, List

from weaverbird.backends.mongo_translator.steps import mongo_step_translator
from weaverbird.pipeline import Pipeline, PipelineStep


def translate_pipeline(
    pipeline_to_translate: Pipeline,
) -> List[Dict[str, Any]]:
    mongo_pipeline = []
    for index, step in enumerate(pipeline_to_translate.steps):
        try:
            mongo_pipeline.extend(mongo_step_translator[step.name](step))
        except Exception as e:
            raise PipelineTranslationFailure(step, index, e) from e
    return mongo_pipeline


class PipelineTranslationFailure(Exception):
    """Raised when a error happens during the translation of the pipeline"""

    def __init__(self, step: PipelineStep, index: int, original_exception: Exception):
        self.step = step
        self.index = index
        self.original_exception = original_exception
        self.message = f'Step #{index + 1} ({step.name}) failed: {original_exception}'
        self.details = {'index': index, 'message': self.message}



from distutils import log

from weaverbird.backends.sql_translator.steps.utils.query_transformation import (
    build_selection_query,
)
from weaverbird.backends.sql_translator.types import (
    SQLDialect,
    SQLPipelineTranslator,
    SQLQuery,
    SQLQueryDescriber,
    SQLQueryExecutor,
    SQLQueryRetriever,
)
from weaverbird.pipeline.steps import ArgminStep


def translate_argmin(
    step: ArgminStep,
    query: SQLQuery,
    index: int,
    sql_query_retriever: SQLQueryRetriever = None,
    sql_query_describer: SQLQueryDescriber = None,
    sql_query_executor: SQLQueryExecutor = None,
    sql_translate_pipeline: SQLPipelineTranslator = None,
    subcall_from_other_pipeline_count: int = None,
    sql_dialect: SQLDialect = None,
) -> SQLQuery:
    query_name = f'ARGMIN_STEP_{index}'

    log.debug(
        '############################################################'
        f'query_name: {query_name}\n'
        '------------------------------------------------------------'
        f'step: {step}\n'
        f'query.transformed_query: {query.transformed_query}\n'
        f'query.metadata_manager.query_metadata: {query.metadata_manager.retrieve_query_metadata()}\n'
    )

    if len(step.groups):
        on_query = "ON " + " AND ".join(
            [f"A.{c}_ALIAS_A=B.{c}" for c in step.groups + [step.column]]
        )
        # the final query
        final_query = (
            f"SELECT * FROM (SELECT {', '.join([f'{c} AS {c}_ALIAS_A' for c in step.groups])},"
            f" MIN({step.column}) AS {step.column}_ALIAS_A FROM {query.query_name}"
            f" GROUP BY {', '.join([f'{c}_ALIAS_A' for c in step.groups])}) A"
            f" INNER JOIN {query.query_name} B"
            f" {on_query}"
        )
    else:
        final_query = (
            f"SELECT * FROM (SELECT MIN({step.column}) AS {step.column}_ALIAS_A FROM {query.query_name}) A"
            f" INNER JOIN {query.query_name} B"
            f" ON A.{step.column}_ALIAS_A=B.{step.column}"
        )
    new_query = SQLQuery(
        query_name=query_name,
        transformed_query=f"{query.transformed_query}, {query_name} AS ({final_query})",
        selection_query=build_selection_query(
            query.metadata_manager.retrieve_query_metadata_columns(), query_name
        ),
        metadata_manager=query.metadata_manager,
    )

    log.debug(
        '------------------------------------------------------------'
        f'SQLquery: {new_query.transformed_query}'
        '############################################################'
    )

    return new_query



from typing import Any, List, Literal, Optional, Union

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName, TemplatedVariable

TYPE_WATERFALL_COLUMN = 'TYPE_waterfall'
LABEL_WATERFALL_COLUMN = 'LABEL_waterfall'
GROUP_WATERFALL_COLUMN = 'GROUP_waterfall'

_RESULT_COLUMN = 'result'


class WaterfallStep(BaseStep):
    name: Literal['waterfall'] = 'waterfall'
    valueColumn: ColumnName
    milestonesColumn: ColumnName
    start: Any
    end: Any
    labelsColumn: ColumnName
    sortBy: Literal['value', 'label']
    order: Literal['desc', 'asc']
    parentsColumn: Optional[ColumnName]
    groupby: List[ColumnName] = []


class WaterfallStepWithVariable(WaterfallStep, StepWithVariablesMixin):
    groupby: Union[TemplatedVariable, List[TemplatedVariable]]



from typing import List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import SplitStep


def translate_split(step: SplitStep) -> List[MongoStep]:
    add_fields_step = {}
    for i in range(1, step.number_cols_to_keep + 1):
        add_fields_step[f'{step.column}_{i}'] = {'$arrayElemAt': ['$_vqbTmp', i - 1]}

    return [
        {'$addFields': {'_vqbTmp': {'$split': [f'${step.column}', step.delimiter]}}},
        {'$addFields': add_fields_step},
        {'$project': {'_vqbTmp': 0}},
    ]



from typing import Any, Dict, List, Union

from pydantic import BaseModel, Field
from typing_extensions import Annotated

from .steps import (
    AddMissingDatesStep,
    AddMissingDatesStepWithVariables,
    AggregateStep,
    AggregateStepWithVariables,
    AppendStep,
    AppendStepWithVariable,
    ArgmaxStep,
    ArgmaxStepWithVariable,
    ArgminStep,
    ArgminStepWithVariable,
    CompareTextStep,
    CompareTextStepWithVariables,
    ConcatenateStep,
    ConcatenateStepWithVariable,
    ConvertStep,
    CumSumStep,
    CumSumStepWithVariable,
    CustomSqlStep,
    CustomStep,
    DateExtractStep,
    DateExtractStepWithVariable,
    DeleteStep,
    DomainStep,
    DuplicateStep,
    DurationStep,
    DurationStepWithVariable,
    EvolutionStep,
    EvolutionStepWithVariable,
    FillnaStep,
    FillnaStepWithVariable,
    FilterStep,
    FilterStepWithVariables,
    FormulaStep,
    FormulaStepWithVariable,
    FromdateStep,
    IfthenelseStep,
    IfThenElseStepWithVariables,
    JoinStep,
    JoinStepWithVariable,
    LowercaseStep,
    MovingAverageStep,
    PercentageStep,
    PivotStep,
    PivotStepWithVariable,
    RankStep,
    RankStepWithVariable,
    RenameStep,
    RenameStepWithVariable,
    ReplaceStep,
    ReplaceStepWithVariable,
    RollupStep,
    RollupStepWithVariable,
    SelectStep,
    SortStep,
    SplitStep,
    SplitStepWithVariable,
    StatisticsStep,
    SubstringStep,
    TextStep,
    TextStepWithVariable,
    ToDateStep,
    TopStep,
    TopStepWithVariables,
    TotalsStep,
    TotalsStepWithVariable,
    TrimStep,
    UniqueGroupsStep,
    UniqueGroupsStepWithVariable,
    UnpivotStep,
    UnpivotStepWithVariable,
    UppercaseStep,
    WaterfallStep,
    WaterfallStepWithVariable,
)

PipelineStep = Annotated[
    Union[
        AddMissingDatesStep,
        AggregateStep,
        AppendStep,
        ArgmaxStep,
        ArgminStep,
        CompareTextStep,
        ConcatenateStep,
        ConvertStep,
        CumSumStep,
        CustomSqlStep,
        CustomStep,
        DateExtractStep,
        DeleteStep,
        DomainStep,
        DuplicateStep,
        DurationStep,
        EvolutionStep,
        FillnaStep,
        FilterStep,
        FormulaStep,
        FromdateStep,
        FromdateStep,
        IfthenelseStep,
        JoinStep,
        LowercaseStep,
        MovingAverageStep,
        PercentageStep,
        PivotStep,
        RankStep,
        RenameStep,
        ReplaceStep,
        RollupStep,
        SelectStep,
        SortStep,
        SplitStep,
        StatisticsStep,
        SubstringStep,
        TextStep,
        ToDateStep,
        TopStep,
        TotalsStep,
        TrimStep,
        UniqueGroupsStep,
        UnpivotStep,
        UppercaseStep,
        WaterfallStep,
    ],
    Field(discriminator='name'),  # noqa: F821
]


class Pipeline(BaseModel):
    steps: List[PipelineStep]

    def dict(self, *, exclude_none: bool = True, **kwargs) -> Dict:
        return super().dict(exclude_none=True, **kwargs)


PipelineStepWithVariables = Annotated[
    Union[
        AddMissingDatesStepWithVariables,
        AggregateStepWithVariables,
        AppendStepWithVariable,
        ArgmaxStepWithVariable,
        ArgminStepWithVariable,
        CompareTextStepWithVariables,
        ConcatenateStepWithVariable,
        CumSumStepWithVariable,
        DateExtractStepWithVariable,
        DurationStepWithVariable,
        EvolutionStepWithVariable,
        FillnaStepWithVariable,
        FilterStepWithVariables,
        FormulaStepWithVariable,
        IfThenElseStepWithVariables,
        JoinStepWithVariable,
        PivotStepWithVariable,
        RankStepWithVariable,
        RenameStepWithVariable,
        ReplaceStepWithVariable,
        RollupStepWithVariable,
        SplitStepWithVariable,
        TextStepWithVariable,
        TopStepWithVariables,
        TotalsStepWithVariable,
        UniqueGroupsStepWithVariable,
        UnpivotStepWithVariable,
        WaterfallStepWithVariable,
    ],
    Field(discriminator='name'),  # noqa: F821
]


class PipelineWithVariables(BaseModel):
    steps: List[Union[PipelineStepWithVariables, PipelineStep]]

    def render(self, variables: Dict[str, Any], renderer) -> Pipeline:
        # TODO it must be more efficient to render the full pipeline once
        steps_rendered = [
            step.render(variables, renderer) if hasattr(step, 'render') else step  # type: ignore
            for step in self.steps
        ]
        return Pipeline(steps=steps_rendered)



from typing import List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import CompareTextStep


def translate_comparetext(step: CompareTextStep) -> List[MongoStep]:
    return [
        {
            '$addFields': {
                (step.new_column_name): {
                    '$cond': [{'$eq': [f'${step.str_col_1}', f'${step.str_col_2}']}, True, False],
                }
            }
        }
    ]



from typing import List, Literal, Union

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName, TemplatedVariable


class ArgmaxStep(BaseStep):
    name: Literal['argmax'] = 'argmax'
    column: ColumnName
    groups: List[str] = []


class ArgmaxStepWithVariable(ArgmaxStep, StepWithVariablesMixin):
    groups: Union[TemplatedVariable, List[TemplatedVariable]]



from typing import List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps.statistics import StatisticsStep


def _get_quantile(n: int, p: int) -> MongoStep:
    """Get n-th p-quantile.

    Examples:
    - the median is the first quantile of order 2.
    - the last decile is the 9-th quantile of order 10.
    """
    return {
        '$avg': [
            {
                '$arrayElemAt': [
                    '$data',
                    {
                        '$trunc': {
                            '$subtract': [{'$multiply': [{'$divide': ['$count', p]}, n]}, 1],
                        },
                    },
                ],
            },
            {
                '$arrayElemAt': [
                    '$data',
                    {
                        '$trunc': {'$multiply': [{'$divide': ['$count', p]}, n]},
                    },
                ],
            },
        ],
    }


def _need_to_compute_column_square(step: StatisticsStep) -> bool:
    return 'variance' in step.statistics or 'standard deviation' in step.statistics


def _need_to_compute_average(step: StatisticsStep) -> bool:
    return (
        'average' in step.statistics
        or 'variance' in step.statistics
        or 'standard deviation' in step.statistics
    )


def _need_to_sort(step: StatisticsStep) -> bool:
    return len(step.quantiles) > 0


def _need_to_count(step: StatisticsStep) -> bool:
    return len(step.quantiles) > 0 or 'count' in step.statistics


# variance = avg(x^2) - avg(x)^2
_VARIANCE_FORMULA = {'$subtract': ['$average_sum_square', {'$pow': ['$average', 2]}]}

_STATISTICS_FORMULA = {
    'count': 1,
    'max': 1,
    'min': 1,
    'average': 1,
    'variance': _VARIANCE_FORMULA,
    'standard deviation': {'$pow': [_VARIANCE_FORMULA, 0.5]},
}


def translate_statistics(step: StatisticsStep) -> List[MongoStep]:
    return [
        {
            '$project': {
                **{col: 1 for col in step.groupby_columns},
                'column': f'${step.column}',
                **(
                    {'column_square': {'$pow': [f'${step.column}', 2]}}
                    if _need_to_compute_column_square(step)
                    else {}
                ),
            },
        },
        {
            '$match': {
                'column': {'$ne': None},
            },
        },
        *([{'$sort': {'column': 1}}] if _need_to_sort(step) else []),
        {
            '$group': {
                '_id': {col: f'${col}' for col in step.groupby_columns} or None,
                'data': {'$push': '$column'},
                **({'count': {'$sum': 1}} if _need_to_count(step) else {}),
                **({'max': {'$max': '$column'}} if 'max' in step.statistics else {}),
                **({'min': {'$min': '$column'}} if 'min' in step.statistics else {}),
                **(
                    {'average_sum_square': {'$avg': '$column_square'}}
                    if _need_to_compute_column_square(step)
                    else {}
                ),
                **({'average': {'$avg': '$column'}} if _need_to_compute_average(step) else {}),
            },
        },
        {
            '$project': {
                #  groupByColumn
                **{col: f'$_id.{col}' for col in step.groupby_columns},
                #  statistics
                **{statistic: _STATISTICS_FORMULA[statistic] for statistic in step.statistics},
                #  quantiles
                **{
                    quantile.label
                    if quantile.label
                    else f'{quantile.nth}-th {quantile.order}-quantile': _get_quantile(
                        quantile.nth, quantile.order
                    )
                    for quantile in step.quantiles
                },
            },
        },
    ]



from pandas import DataFrame

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import ArgmaxStep

_TMP_GROUP_COL_NAME = '__TMP_COL_NAME'


def execute_argmax(
    step: ArgmaxStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    group = step.groups

    if len(step.groups) == 0:
        # if no groups, we create a temp column with a constant in it.
        # Grouping on it should yield a single result
        df[_TMP_GROUP_COL_NAME] = 1
        group = [_TMP_GROUP_COL_NAME]

    aggregated_df = df.groupby(group, as_index=False, dropna=False).agg({step.column: 'max'})

    if len(step.groups) == 0:
        # we now remove the ugly temp column that we grouped on
        del df[_TMP_GROUP_COL_NAME]
        del aggregated_df[_TMP_GROUP_COL_NAME]

    return df.merge(aggregated_df, on=[step.column] + step.groups)



from distutils import log

from weaverbird.backends.sql_translator.steps.utils.query_transformation import (
    apply_condition,
    build_selection_query,
)
from weaverbird.backends.sql_translator.types import (
    SQLDialect,
    SQLPipelineTranslator,
    SQLQuery,
    SQLQueryDescriber,
    SQLQueryExecutor,
    SQLQueryRetriever,
)
from weaverbird.pipeline.steps import IfthenelseStep
from weaverbird.pipeline.steps.ifthenelse import IfThenElse


def build_conditional_expression(step: IfthenelseStep):
    built_string = ''
    current_nested_step = step

    while hasattr(current_nested_step, 'else_value') and isinstance(
        current_nested_step.else_value, IfThenElse
    ):
        built_string += (
            f"WHEN {apply_condition(current_nested_step.condition, '')} "
            f"THEN {current_nested_step.then} "
        )
        current_nested_step = current_nested_step.else_value

    built_string += (
        f"WHEN {apply_condition(current_nested_step.condition, '').replace(',)', ')')} "
        f"THEN {current_nested_step.then} ELSE {current_nested_step.else_value} "
        f"END"
    )
    return built_string.replace('"', "'")


def translate_ifthenelse(
    step: IfthenelseStep,
    query: SQLQuery,
    index: int,
    sql_query_retriever: SQLQueryRetriever = None,
    sql_query_describer: SQLQueryDescriber = None,
    sql_query_executor: SQLQueryExecutor = None,
    sql_translate_pipeline: SQLPipelineTranslator = None,
    subcall_from_other_pipeline_count: int = None,
    sql_dialect: SQLDialect = None,
) -> SQLQuery:
    query_name = f'IFTHENELSE_STEP_{index}'

    log.debug(
        '############################################################'
        f'query_name: {query_name}\n'
        '------------------------------------------------------------'
        f'query.transformed_query: {query.transformed_query}\n'
        f'query.metadata_manager.query_metadata: {query.metadata_manager.retrieve_query_metadata()}\n'
    )

    completed_fields = query.metadata_manager.retrieve_query_metadata_columns_as_str(
        columns_filter=[step.new_column]
    )

    new_query = SQLQuery(
        query_name=query_name,
        transformed_query=f"""{query.transformed_query}, {query_name} AS"""
        f""" (SELECT {completed_fields}, CASE {build_conditional_expression(step)} AS {step.new_column.upper()}"""
        f""" FROM {query.query_name})""",
    )

    query.metadata_manager.add_query_metadata_column(step.new_column, 'str')

    new_query.selection_query = build_selection_query(
        query.metadata_manager.retrieve_query_metadata_columns(), query_name
    )
    new_query.metadata_manager = query.metadata_manager

    log.debug(
        '------------------------------------------------------------'
        f'SQLquery: {new_query.transformed_query}'
        '############################################################'
    )

    return new_query



from typing import Dict, List, Literal

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps.fromdate import FromdateStep

MonthReplacement = Dict[Literal['$switch'], Dict[Literal['branches'], List[MongoStep]]]

_SMALL_MONTH_REPLACE = {
    '$switch': {
        'branches': [
            {'case': {'$eq': ['$_vqbTempMonth', '01']}, 'then': 'Jan'},
            {'case': {'$eq': ['$_vqbTempMonth', '02']}, 'then': 'Feb'},
            {'case': {'$eq': ['$_vqbTempMonth', '03']}, 'then': 'Mar'},
            {'case': {'$eq': ['$_vqbTempMonth', '04']}, 'then': 'Apr'},
            {'case': {'$eq': ['$_vqbTempMonth', '05']}, 'then': 'May'},
            {'case': {'$eq': ['$_vqbTempMonth', '06']}, 'then': 'Jun'},
            {'case': {'$eq': ['$_vqbTempMonth', '07']}, 'then': 'Jul'},
            {'case': {'$eq': ['$_vqbTempMonth', '08']}, 'then': 'Aug'},
            {'case': {'$eq': ['$_vqbTempMonth', '09']}, 'then': 'Sep'},
            {'case': {'$eq': ['$_vqbTempMonth', '10']}, 'then': 'Oct'},
            {'case': {'$eq': ['$_vqbTempMonth', '11']}, 'then': 'Nov'},
            {'case': {'$eq': ['$_vqbTempMonth', '12']}, 'then': 'Dec'},
        ],
    },
}

_FULL_MONTH_REPLACE = {
    '$switch': {
        'branches': [
            {'case': {'$eq': ['$_vqbTempMonth', '01']}, 'then': 'January'},
            {'case': {'$eq': ['$_vqbTempMonth', '02']}, 'then': 'February'},
            {'case': {'$eq': ['$_vqbTempMonth', '03']}, 'then': 'March'},
            {'case': {'$eq': ['$_vqbTempMonth', '04']}, 'then': 'April'},
            {'case': {'$eq': ['$_vqbTempMonth', '05']}, 'then': 'May'},
            {'case': {'$eq': ['$_vqbTempMonth', '06']}, 'then': 'June'},
            {'case': {'$eq': ['$_vqbTempMonth', '07']}, 'then': 'July'},
            {'case': {'$eq': ['$_vqbTempMonth', '08']}, 'then': 'August'},
            {'case': {'$eq': ['$_vqbTempMonth', '09']}, 'then': 'September'},
            {'case': {'$eq': ['$_vqbTempMonth', '10']}, 'then': 'October'},
            {'case': {'$eq': ['$_vqbTempMonth', '11']}, 'then': 'November'},
            {'case': {'$eq': ['$_vqbTempMonth', '12']}, 'then': 'December'},
        ],
    },
}


def _translate_month_year(
    step: FromdateStep, concat_separator: Literal[' ', '-'], month_dict: MongoStep
) -> List[MongoStep]:
    return [
        {
            '$addFields': {
                step.column: {
                    '$dateToString': {'date': f'${step.column}', 'format': '%m-%Y'},
                },
            },
        },
        {'$addFields': {'_vqbTempArray': {'$split': [f'${step.column}', '-']}}},
        {
            '$addFields': {
                '_vqbTempMonth': {'$arrayElemAt': ['$_vqbTempArray', 0]},
                '_vqbTempYear': {'$arrayElemAt': ['$_vqbTempArray', 1]},
            },
        },
        {
            '$addFields': {'_vqbTempMonth': month_dict},
        },
        {
            '$addFields': {
                step.column: {
                    '$concat': ['$_vqbTempMonth', concat_separator, '$_vqbTempYear'],
                },
            },
        },
        {'$project': {'_vqbTempArray': 0, '_vqbTempMonth': 0, '_vqbTempYear': 0}},
    ]


def _translate_day_month_year(
    step: FromdateStep, concat_separator: Literal[' ', '-'], month_dict: MongoStep
) -> List[MongoStep]:
    return [
        {
            '$addFields': {
                step.column: {
                    '$dateToString': {'date': f'${step.column}', 'format': '%d-%m-%Y'},
                },
            },
        },
        {'$addFields': {'_vqbTempArray': {'$split': [f'${step.column}', '-']}}},
        {
            '$addFields': {
                '_vqbTempDay': {'$arrayElemAt': ['$_vqbTempArray', 0]},
                '_vqbTempMonth': {'$arrayElemAt': ['$_vqbTempArray', 1]},
                '_vqbTempYear': {'$arrayElemAt': ['$_vqbTempArray', 2]},
            },
        },
        {
            '$addFields': {'_vqbTempMonth': month_dict},
        },
        {
            '$addFields': {
                step.column: {
                    '$concat': [
                        '$_vqbTempDay',
                        concat_separator,
                        '$_vqbTempMonth',
                        concat_separator,
                        '$_vqbTempYear',
                    ],
                },
            },
        },
        {
            '$project': {
                '_vqbTempArray': 0,
                '_vqbTempDay': 0,
                '_vqbTempMonth': 0,
                '_vqbTempYear': 0,
            }
        },
    ]


def translate_fromdate(step: FromdateStep) -> List[MongoStep]:
    if step.format == '%d %b %Y':
        return _translate_day_month_year(step, ' ', _SMALL_MONTH_REPLACE)
    if step.format == '%d-%b-%Y':
        return _translate_day_month_year(step, '-', _SMALL_MONTH_REPLACE)
    if step.format == '%d %B %Y':
        return _translate_day_month_year(step, ' ', _FULL_MONTH_REPLACE)
    if step.format == '%b %Y':
        return _translate_month_year(step, ' ', _SMALL_MONTH_REPLACE)
    if step.format == '%b-%Y':
        return _translate_month_year(step, '-', _SMALL_MONTH_REPLACE)
    if step.format == '%B %Y':
        return _translate_month_year(step, ' ', _FULL_MONTH_REPLACE)
    else:
        return [
            {
                '$addFields': {
                    step.column: {
                        '$dateToString': {'date': f'${step.column}', 'format': step.format},
                    },
                },
            },
        ]



import json
from typing import List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import CustomStep


def translate_custom(step: CustomStep) -> List[MongoStep]:
    output = json.loads(step.query)
    return output if isinstance(output, list) else [output]



def resolve_sql_pipeline_for_combination(
    pipeline_or_domain,
    sql_query_retriever,
    sql_translate_pipeline,
    sql_query_describer,
    sql_query_executor,
    subcall_from_other_pipeline_count=None,
) -> str:
    """
    Combined pipelines can be either single domains (str), or complete pipeline (list of steps)
    """
    from weaverbird.pipeline import Pipeline

    if isinstance(pipeline_or_domain, str):
        return sql_query_retriever(pipeline_or_domain)
    else:
        # NOTE execution report of the sub-pipeline is discarded
        # For now return the domain name from first step as
        return sql_translate_pipeline(
            Pipeline(steps=pipeline_or_domain),
            sql_query_retriever,
            sql_query_describer,
            sql_query_executor,
            subcall_from_other_pipeline_count,
        )[0]



from contextlib import suppress
from typing import List

from pandas import DataFrame, to_datetime, to_timedelta

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import DateExtractStep
from weaverbird.pipeline.steps.date_extract import DATE_INFO

from .utils.cast import cast_to_int

OPERATIONS_MAPPING = {
    'minutes': 'minute',
    'seconds': 'second',
    'dayOfYear': 'dayofyear',
}


def execute_date_extract(
    step: DateExtractStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    date_info: List[DATE_INFO]
    if step.operation:  # for retrocompatibility
        date_info = [step.operation]
        new_columns = [step.new_column_name or f'{step.column}_{step.operation}']
    else:
        date_info = step.date_info
        new_columns = step.new_columns

    for dt_info, new_col in zip(date_info, new_columns):
        serie_dt = df[step.column].dt
        if dt_info == 'week':
            # cast in float and not in int to manage NaN properly
            result = serie_dt.strftime('%U').astype(float)
        elif dt_info == 'dayOfWeek':
            # result should be between 1 (sunday) and 7 (saturday)
            result = (serie_dt.dayofweek + 2) % 7
            result = result.replace({0: 7})
        elif dt_info == 'isoYear':
            result = serie_dt.isocalendar().year
        elif dt_info == 'isoWeek':
            result = serie_dt.isocalendar().week
        elif dt_info == 'isoDayOfWeek':
            result = serie_dt.isocalendar().day
        elif dt_info == 'firstDayOfYear':
            result = to_datetime(DataFrame({'year': serie_dt.year, 'month': 1, 'day': 1}))
        elif dt_info == 'firstDayOfMonth':
            result = to_datetime(
                DataFrame({'year': serie_dt.year, 'month': serie_dt.month, 'day': 1})
            )
        elif dt_info == 'firstDayOfWeek':
            # dayofweek should be between 1 (sunday) and 7 (saturday)
            dayofweek = (serie_dt.dayofweek + 2) % 7
            dayofweek = dayofweek.replace({0: 7})
            # we subtract a number of days corresponding to(dayOfWeek - 1)
            result = df[step.column] - to_timedelta(dayofweek - 1, unit='d')
            # the result should be returned with 0-ed time information
            result = to_datetime(result.dt.date)
        elif dt_info == 'firstDayOfQuarter':
            result = to_datetime(
                DataFrame(
                    {
                        'year': serie_dt.year,
                        'month': 3 * ((serie_dt.month - 1) // 3) + 1,
                        'day': 1,
                    }
                )
            )
        elif dt_info == 'firstDayOfIsoWeek':
            dayofweek = serie_dt.isocalendar().day
            # we subtract a number of days corresponding to(dayOfWeek - 1)
            result = df[step.column] - to_timedelta(dayofweek - 1, unit='d')
            # the result should be returned with 0-ed time information
            result = to_datetime(result.dt.date)
        elif dt_info == 'previousDay':
            result = df[step.column] - to_timedelta(1, unit='d')
            # the result should be returned with 0-ed time information
            result = to_datetime(result.dt.date)
        elif dt_info == 'firstDayOfPreviousYear':
            result = to_datetime(DataFrame({'year': serie_dt.year - 1, 'month': 1, 'day': 1}))
        elif dt_info == 'firstDayOfPreviousMonth':
            prev_month = serie_dt.month - 1
            prev_month = prev_month.replace({0: 12})
            result = to_datetime(
                DataFrame(
                    {
                        'year': serie_dt.year - (prev_month == 12),
                        'month': prev_month,
                        'day': 1,
                    }
                )
            )
        elif dt_info == 'firstDayOfPreviousWeek':
            prev_week_date = df[step.column] - to_timedelta(7, unit='d')
            # dayofweek should be between 1 (sunday) and 7 (saturday)
            dayofweek = (prev_week_date.dt.dayofweek + 2) % 7
            dayofweek = dayofweek.replace({0: 7})
            # we subtract a number of days corresponding to(dayOfWeek - 1)
            result = prev_week_date - to_timedelta(dayofweek - 1, unit='d')
            # the result should be returned with 0-ed time information
            result = to_datetime(result.dt.date)
        elif dt_info == 'firstDayOfPreviousQuarter':
            first_month_of_quarter = 3 * ((serie_dt.month - 1) // 3) + 1
            first_month_of_prev_q = first_month_of_quarter - 3
            first_month_of_prev_q = first_month_of_prev_q.replace({-2: 10})
            result = to_datetime(
                DataFrame(
                    {
                        'year': serie_dt.year - (first_month_of_prev_q == 10),
                        'month': first_month_of_prev_q,
                        'day': 1,
                    }
                )
            )
        elif dt_info == 'firstDayOfPreviousIsoWeek':
            prev_week_date = df[step.column] - to_timedelta(7, unit='d')
            dayofweek = prev_week_date.dt.isocalendar().day
            # we subtract a number of days corresponding to(dayOfWeek - 1)
            result = prev_week_date - to_timedelta(dayofweek - 1, unit='d')
            # the result should be returned with 0-ed time information
            result = to_datetime(result.dt.date)
        elif dt_info == 'previousYear':
            result = serie_dt.year - 1
        elif dt_info == 'previousMonth':
            month = serie_dt.month
            result = month - 1
            result = result.replace({0: 12})
        elif dt_info == 'previousWeek':
            prev_week_date = df[step.column] - to_timedelta(7, unit='d')
            result = prev_week_date.dt.strftime('%U').astype(float)
        elif dt_info == 'previousQuarter':
            result = serie_dt.quarter - 1
            result = result.replace({0: 4})
        elif dt_info == 'previousIsoWeek':
            prev_week_date = df[step.column] - to_timedelta(7, unit='d')
            result = prev_week_date.dt.isocalendar().week
        elif dt_info == 'milliseconds':
            result = serie_dt.microsecond / 1000
        else:
            operation = OPERATIONS_MAPPING.get(dt_info, dt_info)
            result = getattr(serie_dt, operation)

        # Handle unsigned integers:
        with suppress(AttributeError):
            if result.dtype.is_unsigned_integer:
                result = cast_to_int(result)

        df = df.assign(**{new_col: result})

    return df



from typing import List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import DurationStep

# A mapping of multiplier to apply to convert milliseconds in days, hours, minutes or seconds
DURATION_MULTIPLIER_MAP = {
    'days': 24 * 60 * 60 * 1000,
    'hours': 60 * 60 * 1000,
    'minutes': 60 * 1000,
    'seconds': 1000,
}


def translate_duration(step: DurationStep) -> List[MongoStep]:
    return [
        {
            '$addFields': {
                step.new_column_name: {
                    '$divide': [
                        # a time difference between dates is returned in milliseconds by Mongo
                        {'$subtract': [f'${step.end_date_column}', f'${step.start_date_column}']},
                        DURATION_MULTIPLIER_MAP[step.duration_in],
                    ],
                },
            },
        }
    ]



from distutils import log

from weaverbird.backends.sql_translator.steps.utils.query_transformation import (
    build_selection_query,
)
from weaverbird.backends.sql_translator.types import (
    SQLDialect,
    SQLPipelineTranslator,
    SQLQuery,
    SQLQueryDescriber,
    SQLQueryExecutor,
    SQLQueryRetriever,
)
from weaverbird.pipeline.steps import ConvertStep

PG_BOOLEAN_VALUES = ','.join(
    [
        f'\'{el}\''
        for el in ['t', 'true', 'y', 'yes', 'on', '1', 'f', 'false', 'n', 'no', 'off', '0']
    ]
)


def build_psql(retrieved_col_type, data_type, col):
    if retrieved_col_type == 'TEXT' and data_type == 'boolean':
        return f'CASE WHEN {col} IN ({PG_BOOLEAN_VALUES}) THEN CAST({col} AS {data_type}) ELSE FALSE END AS {col}'
    if retrieved_col_type == 'TEXT' and data_type == 'integer':
        # To ensure we have a unified behavior between pandas, Snowflake & Postgres,
        # additional logic is required to
        # 1. replace any non-numeric valid string by ''
        # 2. Keep the part between '.' (1 is equivalent to 0 in split part
        # see https://docs.snowflake.com/en/sql-reference/functions/split_part.html#arguments)
        # 3. Nullify the output if 1 or 2 results in empty strings i e non numeric value
        # 4. Cast the resulting NULL or string representation of integer as Integer
        return (
            f'CAST(NULLIF(SPLIT_PART(REGEXP_REPLACE({col}, \'[^0-9.]*\', \'\'), \'.\', 1), \'\') '
            f'AS {data_type}) AS {col}'
        )
    if retrieved_col_type == 'FLOAT' and data_type == 'integer':
        return f'TRUNC({col}) AS {col}'

    return f'CAST({col} AS {data_type}) AS {col}'


def build_snowflake(retrieved_col_type, data_type, col):
    if (retrieved_col_type == 'FLOAT' or retrieved_col_type == 'REAL') and data_type == 'integer':
        return f'TRUNCATE({col}) AS {col}'
    elif retrieved_col_type == 'TEXT' and data_type == 'integer':
        return f"CAST(SPLIT_PART({col}, '.', 0) AS {data_type}) AS {col}"
    elif (
        retrieved_col_type == 'TIMESTAMP_NTZ' or retrieved_col_type == 'DATE'
    ) and data_type == 'integer':
        return f"CAST(DATE_PART('EPOCH_MILLISECOND', TO_TIMESTAMP({col})) AS {data_type}) AS {col}"
    return f'CAST({col} AS {data_type}) AS {col}'


def translate_convert(
    step: ConvertStep,
    query: SQLQuery,
    index: int,
    sql_query_retriever: SQLQueryRetriever = None,
    sql_query_describer: SQLQueryDescriber = None,
    sql_query_executor: SQLQueryExecutor = None,
    sql_translate_pipeline: SQLPipelineTranslator = None,
    subcall_from_other_pipeline_count: int = None,
    sql_dialect: SQLDialect = None,
) -> SQLQuery:
    query_name = f'CONVERT_STEP_{index}'

    log.debug(
        '############################################################'
        f'query_name: {query_name}\n'
        '------------------------------------------------------------'
        f'step.columns: {step.columns}\n'
        f'step.data_type: {step.data_type}\n'
        f'query.transformed_query: {query.transformed_query}\n'
        f'query.metadata_manager.query_metadata: {query.metadata_manager.retrieve_query_metadata()}\n'
    )

    is_postgres = sql_dialect == 'postgres'

    to_cast = []
    for col in step.columns:
        retrieved_col_type = query.metadata_manager.retrieve_query_metadata_column_type_by_name(
            column_name=col
        )

        if is_postgres:
            to_cast.append(build_psql(retrieved_col_type, step.data_type, col))
        else:
            to_cast.append(build_snowflake(retrieved_col_type, step.data_type, col))

    for c in step.columns:
        query.metadata_manager.update_query_metadata_column_type(c, step.data_type)

    completed_fields = query.metadata_manager.retrieve_query_metadata_columns_as_str(
        columns_filter=step.columns
    )

    compiled_query = ', '.join(to_cast)
    if len(completed_fields):
        compiled_query = f', {compiled_query}'
    new_query = SQLQuery(
        query_name=query_name,
        transformed_query=f"""{query.transformed_query}, {query_name} AS"""
        f""" (SELECT {completed_fields}{compiled_query}"""
        f""" FROM {query.query_name})""",
        selection_query=build_selection_query(
            query.metadata_manager.retrieve_query_metadata_columns(), query_name
        ),
        metadata_manager=query.metadata_manager,
    )

    log.debug(
        '------------------------------------------------------------'
        f'SQLquery: {new_query.transformed_query}'
        '############################################################'
    )

    return new_query



from typing import Literal

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin


class CustomSqlStep(BaseStep):
    name: Literal['customsql'] = 'customsql'
    query: str


class CustomSqlStepWithVariables(CustomSqlStep, StepWithVariablesMixin):
    ...



from typing import Dict, List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.conditions import (
    ComparisonCondition,
    Condition,
    ConditionComboAnd,
    ConditionComboOr,
    DateBoundCondition,
    InclusionCondition,
    MatchCondition,
    NullCondition,
)
from weaverbird.pipeline.dates import RelativeDate
from weaverbird.pipeline.steps import FilterStep

operator_mapping = {
    'eq': '$eq',
    'ne': '$ne',
    'lt': '$lt',
    'le': '$lte',
    'gt': '$gt',
    'ge': '$gte',
    'in': '$in',
    'nin': '$nin',
    'isnull': '$eq',
    'notnull': '$ne',
    'from': '$gte',
    'until': '$lte',
}

RELATIVE_DATE_OPERATORS: Dict[str, int] = {'until': -1, 'from': +1}


def translate_relative_date(value: RelativeDate):
    sign = RELATIVE_DATE_OPERATORS[value.operator]
    return {
        '$dateAdd': {
            'startDate': truncate_to_day({'$toDate': value.date}),
            'amount': sign * abs(int(value.quantity)),
            'unit': value.duration,
        },
    }


def truncate_to_day(value):
    return {
        '$dateTrunc': {
            'date': value,
            'unit': 'day',
        }
    }


def build_match_tree(condition: Condition, parent_operator='and') -> dict:
    # coumpound conditions

    if isinstance(condition, ConditionComboAnd):
        return {'$and': [build_match_tree(c, 'and') for c in condition.and_]}
    elif isinstance(condition, ConditionComboOr):
        return {'$or': [build_match_tree(c, 'or') for c in condition.or_]}

    # simple conditions
    elif isinstance(condition, ComparisonCondition) or isinstance(condition, InclusionCondition):
        return {condition.column: {operator_mapping[condition.operator]: condition.value}}
    elif isinstance(condition, NullCondition):
        return {condition.column: {operator_mapping[condition.operator]: None}}

    elif isinstance(condition, MatchCondition):
        if condition.operator == 'matches':
            return {condition.column: {'$regex': condition.value}}
        elif condition.operator == 'notmatches':
            return {condition.column: {'$not': {'$regex': condition.value}}}

    # dates
    elif isinstance(condition, DateBoundCondition):
        if isinstance(condition.value, RelativeDate):
            target_date = translate_relative_date(condition.value)
        else:
            target_date = {'$toDate': condition.value}

        return {
            '$expr': {
                operator_mapping[condition.operator]: [
                    truncate_to_day(f'${condition.column}'),
                    truncate_to_day(target_date),
                ]
            }
        }


def translate_filter(step: FilterStep) -> List[MongoStep]:
    return [{'$match': build_match_tree(step.condition)}]



from typing import List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps.fillna import FillnaStep


def translate_fillna(step: FillnaStep) -> List[MongoStep]:
    add_fields = {col: {'$ifNull': [f'${col}', step.value]} for col in step.columns}
    return [{'$addFields': add_fields}]



from typing import List, Literal

from pydantic import BaseModel

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.types import ColumnName, PopulatedWithFieldnames


class ColumnSort(BaseModel):
    class Config(PopulatedWithFieldnames):
        ...

    column: ColumnName
    order: Literal['asc', 'desc']


class SortStep(BaseStep):
    name: Literal['sort'] = 'sort'
    columns: List[ColumnSort]



from abc import ABC
from datetime import datetime
from typing import Any, List, Literal, Union

from pydantic import BaseModel, Field
from typing_extensions import Annotated

from weaverbird.pipeline.dates import RelativeDate
from weaverbird.pipeline.types import ColumnName, PopulatedWithFieldnames


class BaseCondition(BaseModel):
    ...


class ComparisonCondition(BaseCondition):
    column: ColumnName
    operator: Literal['eq', 'ne', 'lt', 'le', 'gt', 'ge']
    value: Any


class InclusionCondition(BaseCondition):
    column: ColumnName
    operator: Literal['in', 'nin']
    value: List[Any]


class NullCondition(BaseCondition):
    column: ColumnName
    operator: Literal['isnull', 'notnull']


class MatchCondition(BaseCondition):
    column: ColumnName
    operator: Literal['matches', 'notmatches']
    value: str


class DateBoundCondition(BaseModel):
    column: ColumnName
    operator: Literal['from', 'until']
    value: Union[RelativeDate, datetime, str]


SimpleCondition = Annotated[
    Union[
        ComparisonCondition,
        InclusionCondition,
        NullCondition,
        MatchCondition,
        DateBoundCondition,
    ],
    Field(discriminator='operator'),  # noqa: F821
]


class BaseConditionCombo(BaseCondition, ABC):
    class Config(PopulatedWithFieldnames):
        ...

    def to_dict(self):
        return self.dict(by_alias=True)


class ConditionComboAnd(BaseConditionCombo):
    and_: List['Condition'] = Field(..., alias='and')


class ConditionComboOr(BaseConditionCombo):
    or_: List['Condition'] = Field(..., alias='or')


Condition = Union[ConditionComboAnd, ConditionComboOr, SimpleCondition]
ConditionComboOr.update_forward_refs()
ConditionComboAnd.update_forward_refs()



from typing import Any, Callable, Dict

from .base import BaseStep


class StepWithVariablesMixin:
    def render(self, variables: Dict[str, Any], renderer: Callable[[Any, Any], Any]) -> BaseStep:
        RenderedClass = self.__class__.__bases__[0]
        step_as_dict = self.dict()  # type: ignore
        rendered_dict = renderer(step_as_dict, variables)
        return RenderedClass(**rendered_dict)



from typing import List, Literal

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.types import ColumnName


class DeleteStep(BaseStep):
    name: Literal['delete'] = 'delete'
    columns: List[ColumnName]



from distutils import log

from weaverbird.backends.sql_translator.steps.utils.query_transformation import (
    build_selection_query,
)
from weaverbird.backends.sql_translator.types import (
    SQLDialect,
    SQLPipelineTranslator,
    SQLQuery,
    SQLQueryDescriber,
    SQLQueryExecutor,
    SQLQueryRetriever,
)
from weaverbird.pipeline.steps import CustomSqlStep


def translate_customsql(
    step: CustomSqlStep,
    query: SQLQuery,
    index: int,
    sql_query_retriever: SQLQueryRetriever = None,
    sql_query_describer: SQLQueryDescriber = None,
    sql_query_executor: SQLQueryExecutor = None,
    sql_translate_pipeline: SQLPipelineTranslator = None,
    subcall_from_other_pipeline_count: int = None,
    sql_dialect: SQLDialect = None,
) -> SQLQuery:
    query_name = f'CUSTOMSQL_STEP_{index}'

    log.debug(
        '############################################################'
        f'query_name: {query_name}\n'
        '------------------------------------------------------------'
        f'step: {step}\n'
        f'query.transformed_query: {query.transformed_query}\n'
        f'query.metadata_manager.query_metadata: {query.metadata_manager.retrieve_query_metadata()}\n'
    )
    injected_query = (
        step.query.replace('##PREVIOUS_STEP##', query.query_name).replace(';', '').strip()
    )
    transformed_query = f"""{query.transformed_query}, {query_name} AS ({injected_query})"""
    query.metadata_manager.remove_query_metadata_all_columns()
    query.metadata_manager.add_query_metadata_columns(
        sql_query_describer(
            domain=None, query_string=f'{transformed_query} SELECT * FROM {query_name}'
        )
    )

    return SQLQuery(
        query_name=query_name,
        transformed_query=transformed_query,
        selection_query=build_selection_query(
            query.metadata_manager.retrieve_query_metadata_columns(), query_name
        ),
        metadata_manager=query.metadata_manager,
    )



from typing import Literal

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.types import ColumnName


class UppercaseStep(BaseStep):
    name: Literal['uppercase'] = 'uppercase'
    column: ColumnName



import time


class StopWatch:
    def __enter__(self):
        self.begin = time.perf_counter()

    def __exit__(self, type, value, traceback):
        self.end = time.perf_counter()
        self.interval = self.end - self.begin



from typing import Dict

from ..types import SQLStepTranslator
from .aggregate import translate_aggregate
from .append import translate_append
from .argmax import translate_argmax
from .argmin import translate_argmin
from .concatenate import translate_concatenate
from .convert import translate_convert
from .cumsum import translate_cumsum
from .customsql import translate_customsql
from .dateextract import translate_dateextract
from .delete import translate_delete
from .duplicate import translate_duplicate
from .duration import translate_duration
from .evolution import translate_evolution
from .fillna import translate_fillna
from .filter import translate_filter
from .formula import translate_formula
from .fromdate import translate_fromdate
from .ifthenelse import translate_ifthenelse
from .join import translate_join
from .lowercase import translate_lowercase
from .percentage import translate_percentage
from .pivot import translate_pivot
from .rank import translate_rank
from .rename import translate_rename
from .replace import translate_replace
from .rollup import translate_rollup
from .select import translate_select
from .sort import translate_sort
from .split import translate_split
from .substring import translate_substring
from .table import translate_table
from .text import translate_text
from .todate import translate_todate
from .top import translate_top
from .totals import translate_totals
from .uniquegroups import translate_uniquegroups
from .unpivot import translate_unpivot
from .uppercase import translate_uppercase

sql_steps_translators: Dict[str, SQLStepTranslator] = {
    'domain': translate_table,  # type ignore # TODO to update
    'filter': translate_filter,
    'aggregate': translate_aggregate,
    'select': translate_select,
    'ifthenelse': translate_ifthenelse,
    'sort': translate_sort,
    'rename': translate_rename,
    'convert': translate_convert,
    'text': translate_text,
    'lowercase': translate_lowercase,
    'uppercase': translate_uppercase,
    'fromdate': translate_fromdate,
    'todate': translate_todate,
    'formula': translate_formula,
    'replace': translate_replace,
    'join': translate_join,
    'uniquegroups': translate_uniquegroups,
    'top': translate_top,
    'percentage': translate_percentage,
    'unpivot': translate_unpivot,
    'pivot': translate_pivot,
    'dateextract': translate_dateextract,
    'append': translate_append,
    'concatenate': translate_concatenate,
    'delete': translate_delete,
    'rank': translate_rank,
    'split': translate_split,
    'substring': translate_substring,
    'fillna': translate_fillna,
    'customsql': translate_customsql,
    'argmax': translate_argmax,
    'argmin': translate_argmin,
    'duplicate': translate_duplicate,
    'evolution': translate_evolution,
    'duration': translate_duration,
    'rollup': translate_rollup,
    'cumsum': translate_cumsum,
    'totals': translate_totals,
}



from typing import Any, Callable, List, Optional, Protocol, Tuple, Union

from pandas import DataFrame
from pydantic import BaseModel, Field

from weaverbird.pipeline import Pipeline
from weaverbird.pipeline.steps.utils.combination import Reference


class StepExecutionReport(BaseModel):
    step_index: int
    time_spent_in_ms: int
    memory_used_in_bytes: int


class PipelineExecutionReport(BaseModel):
    steps_reports: List[StepExecutionReport] = Field(min_items=0)


DomainRetriever = Callable[[Union[str, Reference]], DataFrame]
PipelineExecutor = Callable[[Pipeline, DomainRetriever], Tuple[DataFrame, PipelineExecutionReport]]


class StepExecutor(Protocol):
    def __call__(
        self,
        step: Any,
        df: DataFrame,
        domain_retriever: Optional[DomainRetriever],
        execute_pipeline: Optional[PipelineExecutor],
    ) -> DataFrame:
        ...



from distutils import log

from weaverbird.backends.sql_translator.steps.utils.combination import (
    resolve_sql_pipeline_for_combination,
)
from weaverbird.backends.sql_translator.steps.utils.query_transformation import (
    build_selection_query,
)
from weaverbird.backends.sql_translator.types import (
    SQLDialect,
    SQLPipelineTranslator,
    SQLQuery,
    SQLQueryDescriber,
    SQLQueryExecutor,
    SQLQueryRetriever,
)
from weaverbird.pipeline.steps import JoinStep


def translate_join(
    step: JoinStep,
    query: SQLQuery,
    index: int,
    sql_query_retriever: SQLQueryRetriever = None,
    sql_query_describer: SQLQueryDescriber = None,
    sql_query_executor: SQLQueryExecutor = None,
    sql_translate_pipeline: SQLPipelineTranslator = None,
    subcall_from_other_pipeline_count: int = None,
    sql_dialect: SQLDialect = None,
) -> SQLQuery:
    query_name = f'JOIN_STEP_{index}'
    log.debug(
        '############################################################'
        f'query_name: {query_name}\n'
        '------------------------------------------------------------'
        f'step: {step}\n'
        f'query.transformed_query: {query.transformed_query}\n'
        f'query.metadata_manager.query_metadata: {query.metadata_manager.retrieve_query_metadata()}\n'
    )

    if step.type == 'left outer':
        how = 'LEFT'
    else:
        how = step.type.upper()

    # Retrieve the right query either directly as a domain or as a resolved sql pipeline
    right_query = resolve_sql_pipeline_for_combination(
        step.right_pipeline,
        sql_query_retriever,
        sql_translate_pipeline,
        sql_query_describer,
        sql_query_executor,
    )
    right_query_name = f'JOIN_STEP_{index}_RIGHT'
    # Update tables metadata with joined table metadata

    # 1 retrieve right metadata
    query_to_join_metadata = sql_query_describer(domain=None, query_string=right_query)

    # 2 Add them to the MetadataManager
    query.metadata_manager.create_table(right_query_name)
    query.metadata_manager.add_table_columns_from_dict(right_query_name, query_to_join_metadata)

    # 3 Update transformed query string with right query selection
    transformed_query = f'{query.transformed_query}, {right_query_name} AS ({right_query})'

    # 4 Join left & right metadata in internal metadata
    query.metadata_manager.join_query_metadata(right_query_name, left_query_name=query.query_name)

    # 5 build the final query string
    join_part = f"{'AND'.join([f'{query.query_name}.{keys[0]} = {right_query_name}.{keys[1]}' for keys in step.on])}"

    transformed_query = f"""{transformed_query}, {query_name} AS (SELECT {query.metadata_manager.retrieve_query_metadata_columns_as_str()} FROM {query.query_name} {how} JOIN {right_query_name} ON {join_part})"""

    log.debug(
        '------------------------------------------------------------'
        f'SQLquery: {transformed_query}'
        '############################################################'
    )

    new_query = SQLQuery(
        query_name=query_name,
        transformed_query=transformed_query,
        selection_query=build_selection_query(
            query.metadata_manager.retrieve_query_metadata_columns(), query_name
        ),
        metadata_manager=query.metadata_manager,
    )

    query.metadata_manager.update_query_metadata_column_names_with_alias()
    return new_query



from distutils import log

from weaverbird.backends.sql_translator.steps.utils.query_transformation import (
    build_selection_query,
    sanitize_column_name,
)
from weaverbird.backends.sql_translator.types import (
    SQLDialect,
    SQLPipelineTranslator,
    SQLQuery,
    SQLQueryDescriber,
    SQLQueryExecutor,
    SQLQueryRetriever,
)
from weaverbird.pipeline.steps import PivotStep


def translate_pivot(
    step: PivotStep,
    query: SQLQuery,
    index: int,
    sql_query_retriever: SQLQueryRetriever = None,
    sql_query_describer: SQLQueryDescriber = None,
    sql_query_executor: SQLQueryExecutor = None,
    sql_translate_pipeline: SQLPipelineTranslator = None,
    subcall_from_other_pipeline_count: int = None,
    sql_dialect: SQLDialect = None,
) -> SQLQuery:
    query_name = f'PIVOT_STEP_{index}'

    log.debug(
        '############################################################'
        f'query_name: {query_name}\n'
        '------------------------------------------------------------'
        f'step: {step}\n'
        f'query.transformed_query: {query.transformed_query}\n'
        f'query.metadata_manager.query_metadata: {query.metadata_manager.retrieve_query_metadata()}\n'
    )
    aggregate_part = f'{step.agg_function}({step.value_column})'
    pivot_values = sql_query_executor(
        domain=None,
        query_string=f"""{query.transformed_query} SELECT DISTINCT({step.column_to_pivot}) FROM {query.query_name}""",
    )[f'{step.column_to_pivot}'].values.tolist()
    sanitized_columns = [sanitize_column_name(p) for p in pivot_values]

    pivoted_values_column_type = query.metadata_manager.retrieve_query_metadata_column_type_by_name(
        step.value_column
    )
    query.metadata_manager.remove_query_metadata_columns(
        query.metadata_manager.retrieve_query_metadata_columns_as_list(columns_filter=step.index)
    )
    query.metadata_manager.add_query_metadata_columns(
        columns={name: pivoted_values_column_type for name in sanitized_columns}
    )

    prepivot_query = (
        f'''PRE_{query_name} AS (SELECT {', '.join(step.index + [step.column_to_pivot, step.value_column])} FROM'''
        f''' {query.query_name})'''
    )
    pivot_query = (
        f"""SELECT {query.metadata_manager.retrieve_query_metadata_columns_as_str()}"""
        f""" FROM PRE_{query_name} PIVOT({aggregate_part} FOR {step.column_to_pivot} IN """
        f"""({', '.join([f"'{val}'" for val in pivot_values])})) AS p ({', '.join(step.index + sanitized_columns)})"""
    )
    transformed_query = (
        f"""{query.transformed_query}, {prepivot_query}, {query_name} AS ({pivot_query})"""
    )

    new_query = SQLQuery(
        query_name=query_name,
        transformed_query=transformed_query,
        selection_query=build_selection_query(
            query.metadata_manager.retrieve_query_metadata_columns(), query_name
        ),
        metadata_manager=query.metadata_manager,
    )

    log.debug(
        '------------------------------------------------------------'
        f'SQLquery: {new_query.transformed_query}'
        '############################################################'
    )

    query.metadata_manager.update_query_metadata_column_names_with_alias()
    return new_query



from weaverbird.backends.sql_translator.steps.utils.query_transformation import (
    build_selection_query,
    sanitize_input,
)
from weaverbird.backends.sql_translator.types import (
    SQLDialect,
    SQLPipelineTranslator,
    SQLQuery,
    SQLQueryDescriber,
    SQLQueryExecutor,
    SQLQueryRetriever,
)
from weaverbird.pipeline.steps import FillnaStep


def translate_fillna(
    step: FillnaStep,
    query: SQLQuery,
    index: int,
    sql_query_retriever: SQLQueryRetriever = None,
    sql_query_describer: SQLQueryDescriber = None,
    sql_query_executor: SQLQueryExecutor = None,
    sql_translate_pipeline: SQLPipelineTranslator = None,
    subcall_from_other_pipeline_count: int = None,
    sql_dialect: SQLDialect = None,
) -> SQLQuery:
    query_name = f'FILLNA_STEP_{index}'

    unchanged_colums = query.metadata_manager.retrieve_query_metadata_columns_as_str(
        columns_filter=step.columns
    )
    fill_value = f"'{sanitize_input(step.value)}'" if isinstance(step.value, str) else step.value
    filled_columns = ', '.join([f'COALESCE({col}, {fill_value}) AS {col}' for col in step.columns])
    new_query = SQLQuery(
        query_name=query_name,
        transformed_query=f"""{query.transformed_query}, {query_name} AS (SELECT {unchanged_colums}, {filled_columns}\
 FROM {query.query_name})""",
        selection_query=build_selection_query(
            query.metadata_manager.retrieve_query_metadata_columns(), query_name
        ),
        metadata_manager=query.metadata_manager,
    )
    return new_query
